#include <cstddef>
#include <cstdint>
#include <cstdio>
#include <cstdlib>
#include <zip.h>

// Forward-declare the libFuzzer's mutator callback.
extern "C" size_t
LLVMFuzzerMutate(uint8_t *Data, size_t Size, size_t MaxSize);

// The custom mutator:
//   * deserialize the data (in this case, uncompress).
//     * If the data doesn't deserialize, create a properly serialized dummy.
//   * Mutate the deserialized data (in this case, just call LLVMFuzzerMutate).
//   * Serialize the mutated data (in this case, compress).
extern "C" size_t LLVMFuzzerCustomMutator(uint8_t *Data, size_t Size,
                                          size_t MaxSize, unsigned int Seed) {

  zip_error_t* error;
  zip_source_t* source = zip_source_buffer_create(Data, Size, 1, error);
  if(source == NULL) {
    printf("Could not create zip source from buffer\n");
    return 0;
  }

  zip_t* archive = zip_open_from_source(source, 0, error);
  if(archive == NULL) {
    printf("Could not open zip archive\n");
    return 0;
  }

  struct zip_stat stats;
  zip_stat_init(&stats);
  zip_source_stat(source, &stats);
  char *uncompressed = new char[stats.size];
  if(zip_source_read(source, uncompressed, stats.size) == -1) {
    printf("Could not read zip archive\n");
    return 0;
  }

  size_t uncompressedLen = LLVMFuzzerMutate((uint8_t*)uncompressed, stats.size, sizeof(uncompressed));

  zip_source_begin_write(source);
  if(zip_source_write(source, uncompressed, uncompressedLen) == -1) {
    printf("Could not write zip archive\n");
    return 0;
  }
  zip_source_commit_write(source);
  zip_source_stat(source, &stats);

  zip_source_close(source);
  delete uncompressed;

  return stats.comp_size;

//  uint8_t Uncompressed[100];
//  size_t UncompressedLen = sizeof(Uncompressed);
//  size_t CompressedLen = MaxSize;
//  if (Z_OK != uncompress(Uncompressed, &UncompressedLen, Data, Size)) {
//    // The data didn't uncompress.
//    // So, it's either a broken input and we want to ignore it,
//    // or we've started fuzzing from an empty corpus and we need to supply
//    // out first properly compressed input.
//    uint8_t Dummy[] = {'H', 'i'};
//    if (Z_OK != compress(Data, &CompressedLen, Dummy, sizeof(Dummy)))
//      return 0;
//    // fprintf(stderr, "Dummy: max %zd res %zd\n", MaxSize, CompressedLen);
//    return CompressedLen;
//  }
//  UncompressedLen =
//      LLVMFuzzerMutate(Uncompressed, UncompressedLen, sizeof(Uncompressed));
//  if (Z_OK != compress(Data, &CompressedLen, Uncompressed, UncompressedLen))
//    return 0;
//  return CompressedLen;
}